#ifndef __DH_H__
#define __DH_H__

/**
 * Functions useful for calculating the forward kinematics of a robot.
 * Some is taken from the Motion Strategy Library and some is taken from 
 * the book by Spong, Hutchinson & Vidyasager p. 77. 
 */
#include <ompl/config.h>

#include <math.h>
#include <vector>

#ifndef sqr
#define sqr(x) (x*x)
#endif

#ifndef PI
#define PI 3.14159265
#endif

namespace ob = ompl::base;
class DH
{
 public:
  /**
   * Calculates transformation matrix from Denavit-Hartenberg parameters.
   * Taken from Spong, Hutchinson & Vidyasager p. 77. 
   */
static ob::ProjectionMatrix::Matrix makeAmatrix(double D, double Theta, double A, double Alpha) {
  	ob::ProjectionMatrix::Matrix r(4,4);
    r(0,0) = cos(Theta);
    r(1,0) = sin(Theta);
    r(2,0) = 0;
    r(3,0) = 0;
    r(0,1) = -sin(Theta) * cos(Alpha);
    r(1,1) = cos(Theta) * cos(Alpha);
    r(2,1) = sin(Alpha);
    r(3,1) = 0;
    r(0,2) = sin(Theta)*sin(Alpha);
    r(1,2) = -cos(Theta)*sin(Alpha);
    r(2,2) = cos(Alpha);
    r(3,2) = 0;
    r(0,3) = A * cos(Theta);
    r(1,3) = A * sin(Theta);
    r(2,3) = D;
    r(3,3) = 1.0;
    return r;
}
  
static void transpose(ob::ProjectionMatrix::Matrix &matrix) {
    for (int i = 0; i < 4; i++) {
    	for (int j = i + 1; j < 4; j++) {
			int save = matrix(i,j);
			matrix(i,j) = matrix(j,i);
			matrix(j,i) = save;
   		}
    }
}
/*static ob::ProjectionMatrix::Matrix inverse(const ob::ProjectionMatrix::Matrix &m) {
  	ob::ProjectionMatrix::Matrix R(3,3);
    R(0,0) = m(0,0);
    R(0,1) = m(0,1);
    R(0,2) = m(0,2);
    R(1,0) = m(1,0);
    R(1,1) = m(1,1);
    R(1,2) = m(1,2);
    R(2,0) = m(2,0);
    R(2,1) = m(2,1);
    R(2,2) = m(2,2);
    std::vector<double> d(3);
    d[0] = m(0,3);
    d[1] = m(1,3);
    d[2] = m(2,3);
    ob::ProjectionMatrix::Matrix result(4,4);
    //Transpose R
    DH::transpose(R);
    d = -R*d;
    result(0,0) = R(0,0);
    result(0,1) = R(0,1);
    result(0,2) = R(0,2);
    result(1,0) = R(1,0);
    result(1,1) = R(1,1);
    result(1,2) = R(1,2);
    result(2,0) = R(2,0);
    result(2,1) = R(2,1);
    result(2,2) = R(2,2);
    result(0,3) = d[0];
    result(1,3) = d[1];
    result(2,3) = d[2];
    result(3,3) = 1.0;
    result(3,0) = 0.0f;
    result(3,1) = 0.0f;
    result(3,2) = 0.0f;
    return result;
  }*/

  /**
   * Multiply two matrices and store the result in r
   * r = ro * rn
   */
  static void mul(ob::ProjectionMatrix::Matrix &r, ob::ProjectionMatrix::Matrix &ro, const ob::ProjectionMatrix::Matrix &rn) {
    for (int k=0; k<4; k++){
      for (int l=0; l<4; l++) {
		r(k,l)=0.0 ;
		for (int m=0; m<4; m++){
		  r(k,l)+=ro(k,m)*rn(m,l);
		}
      }
    }
  }

//Calculate full DH for a robot with 6 links
static void fullDHIRB140(ob::ProjectionMatrix::Matrix &r, ob::ProjectionMatrix::Matrix &a, ob::ProjectionMatrix::Matrix &b, ob::ProjectionMatrix::Matrix &c, ob::ProjectionMatrix::Matrix &d, ob::ProjectionMatrix::Matrix &e, ob::ProjectionMatrix::Matrix &f){
	ob::ProjectionMatrix::Matrix r2(4,4);
	DH::mul(r, a, b);	
	
	/*std::cout << "*********** A1A2 ***********" << std::endl;
	for(int a=0; a<4;a++)
	{
		for(int b=0; b<4; b++)
		{
			std::cout << r(a,b) << "\t";
		}
		std::cout << std::endl;
	}	*/
	DH::mul(r2, r, c);
	/*std::cout << "*********** A3 ***********" << std::endl;
	for(int a=0; a<4;a++)
	{
		for(int b=0; b<4; b++)
		{
			std::cout << r2(a,b) << "\t";
		}
		std::cout << std::endl;
	}	*/
	DH::mul(r, r2, d);
	/*std::cout << "*********** A4 ***********" << std::endl;
	for(int a=0; a<4;a++)
	{
		for(int b=0; b<4; b++)
		{
			std::cout << r(a,b) << "\t";
		}
		std::cout << std::endl;
	}	*/
	DH::mul(r2, r, e);
	/*std::cout << "*********** A5 ***********" << std::endl;
	for(int a=0; a<4;a++)
	{
		for(int b=0; b<4; b++)
		{
			std::cout << r2(a,b) << "\t";
		}
		std::cout << std::endl;
	}	*/
	DH::mul(r, r2, f);
	/*std::cout << "*********** A5 ***********" << std::endl;
	for(int a=0; a<4;a++)
	{
		for(int b=0; b<4; b++)
		{
			std::cout << r(a,b) << "\t";
		}
		std::cout << std::endl;
	}	*/
}
  /**
   * Extract position and orientation from a transformation matrix r.
   * The information is stored in the vector q at offset offset.
   */
static void extract(std::vector<double> &q, unsigned int offset, ob::ProjectionMatrix::Matrix &r) {
    q[offset] = r(0,3); 
    q[offset+1] = r(1,3);
    q[offset+2] = r(2,3);

    q[offset+4] = atan2(-r(2,0), sqrt(sqr(r(0,0))+sqr(r(1,0))));
    if ((q[offset+4] > 0) && (cos(q[offset+4])==0.0)) {
      q[offset+5] = atan2(r(1,0), r(0,0));
      q[offset+3] = 0.0f;
    } else if ((q[offset+4] < 0) && (cos(q[offset+4])==0.0)) {
      q[offset+5] = -atan2(r(1,0), r(0,0));
      q[offset+3] = 0.0f;
    } else {
      q[offset+5] = atan2(r(1,0), r(0,0));
      q[offset+3] = atan2(r(2,1), r(2,2));
    }
	// Force the orientation into [-pi,pi)
    if (q[offset+3] < -PI){
      q[offset+3] += 2.0*PI; 
     }
    else if (q[offset+3] >= PI){
      q[offset+3] -= 2.0*PI;
    }
    if (q[offset+4] < -PI){
      q[offset+4] += 2.0*PI; 
    }
    else if (q[offset+4] >= PI){
      q[offset+4] -= 2.0*PI;
    }
    if (q[offset+5] < -PI){
      q[offset+5] += 2.0*PI;
    }
    else if (q[offset+5] >= PI){
      q[offset+5] -= 2.0*PI;
    }
  }

static std::vector<double> inverseKinematics(std::vector<double> pos, int offset, ob::ProjectionMatrix::Matrix DH) {
	std::vector<double> r(6);
	float xc = pos[0] - 65*DH(0,2) - 70*DH(0,2);
	float yc = pos[1] - 65*DH(1,2) - 70*DH(1,2);
	float zc = pos[2] - 65*DH(2,2);
	std::cout << "+++++++++++++ inverse +++++++++++++" << std::endl;
	//This is hardcoded for the ABB IRB140 atm. With fixed d's and a's
	//Theta1
/*	r[0] = atan2(yc, xc) - atan2(offset, sqrt((xc*xc) + (yc*yc) - (offset*offset)));
	
	//Theta3
	float D = ((xc*xc) + (yc*yc) - (offset*offset) + ((zc -352)*(zc - 352)) -(360*360) - (380*380))/(2*360*380);
	r[2] = -PI/2;//atan2(sqrt(1-(D*D)),D);
	
	//Theta2
	r[1] = 	atan2(zc - 352, sqrt(xc*xc + yc*yc -offset*offset)) - atan2(380*sin(r[2]), 360+380*cos(r[2]));
	
	//Theta4
	r[3] = atan2(cos(r[0])*cos(r[1]+r[2])*DH(0,2) + sin(r[0])*cos(r[1]+r[2])*DH(1,2) + sin(r[1]+r[2])*DH(2,2), -cos(r[0])*sin(r[1]+r[2])*DH(0,2) - sin(r[0])*sin(r[1]+r[2])*DH(1,2) + cos(r[1]+r[2])*DH(2,2));
	
	//Theta5
	r[4] = atan2(sin(r[0])*DH(0,2) - cos(r[0])*DH(1,2), sqrt(1- (sin(r[0])*DH(0,2) - cos(r[0])*DH(1,2))*(sin(r[0])*DH(0,2) - cos(r[0])*DH(1,2))));
	
	//Theta6
	r[5] = atan2(-sin(r[0])*DH(0,0) + cos(r[0])*DH(1,0), sin(r[0])*DH(0,1) - cos(r[0])*DH(1,1));*/
	
	//NO offset
	
	r[0] = atan2(yc,xc);// + atan2(-offset, -sqrt((xc*xc) + (yc*yc) - (offset*offset)));;
	
	float D = (xc*xc + yc*yc - (offset*offset) + (zc-352)*(zc-352) - 360*360 - 380*380) / (2*360*380);
	r[2] = atan2(-sqrt(1-D*D), D);
	
	r[1] = atan2(zc -352, sqrt(xc*xc + yc*yc - offset*offset)) - atan2(380*sin(r[2]), 360 + 380*cos(r[2]));
	
	r[3] = atan2(	-cos(r[0])*sin(r[1]+r[2])*DH(0,2) - sin(r[0])*sin(r[1]+r[2])*DH(1,2) + cos(r[1]+r[2])*DH(2,2),
					cos(r[0])*cos(r[1]+r[2])*DH(0,2) + sin(r[0])*cos(r[1]+r[2])*DH(1,2) + sin(r[1]+r[2])*DH(2,2));
	std::cout << "r[4] = atan2("<<sin(r[0])*DH(0,2) - cos(r[0])*DH(1,2)<<","<<-sqrt(1-(sin(r[0])*DH(0,2) - cos(r[0])*DH(1,2))*(sin(r[0])*DH(0,2) - cos(r[0])*DH(1,2)))<<")"<<std::endl;
	std::cout << "r[5] = atan2("<<sin(r[0])*DH(0,1) - cos(r[0])*DH(1,1)<<","<<					-sin(r[0])*DH(0,0) + cos(r[0])*DH(1,0)<<")"<<std::endl;
	r[4] = atan2(	sin(r[0])*DH(0,2) - cos(r[0])*DH(1,2),
					-sqrt(1-(sin(r[0])*DH(0,2) - cos(r[0])*DH(1,2))*(sin(r[0])*DH(0,2) - cos(r[0])*DH(1,2)))
					);
	r[5] = atan2(	sin(r[0])*DH(0,1) - cos(r[0])*DH(1,1),
					-sin(r[0])*DH(0,0) + cos(r[0])*DH(1,0));
	
	for (int i=0;i<6;i++) {
		if (r[i] >= PI) {
			r[i] = r[i] -PI;
		}
		else if (r[i] <= -PI) {
			r[i] = r[i] + PI;
		}
	}
	
	return r;
}
  static void identity(ob::ProjectionMatrix::Matrix &r) {
    for (int i = 0; i < 4 ; i ++ ) {
      for (int j = 0 ; j < 4 ; j ++ ) {
		if (i==j) {
				r(i,j) = 1.0;
		}
		else {
			r(i,j) = 0.0;
		}
      }
    }
  }

  static void zeros(ob::ProjectionMatrix::Matrix &r) {
    for (int i = 0; i < 4 ; i ++ ) {
      for (int j = 0 ; j < 4 ; j ++ ) {
	r(i,j) = 0.0;
      }
    }
  }
};

#endif
