#include "myStateValidityCheckerClass.h"
#include <vector>
#include <typeinfo>

namespace ob = ompl::base;

myStateValidityCheckerClass::myStateValidityCheckerClass(const ob::SpaceInformationPtr &si) : ob::StateValidityChecker(si)
{
	loadCornerPoints();
	environmentTriangles = importEnvironment("irb/environment", myStateValidityCheckerClass::environment, 0);
}	

bool myStateValidityCheckerClass::isValid(const ob::State *state) const
{
	//Cast abstract state to the one we expect
	const ob::RealVectorStateSpace::StateType *state1 = state->as<ob::RealVectorStateSpace::StateType>();
	
	//Calculate DH transformations for current state
	ob::ProjectionMatrix::Matrix A0= DH::makeAmatrix(0, state1->values[0], 0, 0);
	
	ob::ProjectionMatrix::Matrix A1 = DH::makeAmatrix(352, state1->values[0], 70, -PI/2);
	ob::ProjectionMatrix::Matrix A2 = DH::makeAmatrix(0, state1->values[1] + (-PI/2), 360, 0);
	ob::ProjectionMatrix::Matrix A3 = DH::makeAmatrix(0, state1->values[2] + 0, 0, -PI/2);
	ob::ProjectionMatrix::Matrix A4 = DH::makeAmatrix(445, state1->values[3] + 0, 0, PI/2);
	ob::ProjectionMatrix::Matrix A5 = DH::makeAmatrix(0, state1->values[4] + 0, 0, -PI/2);
	ob::ProjectionMatrix::Matrix A6 = DH::makeAmatrix(0, state1->values[5] + PI, 0, 0);
	
	//COLLISION DETECTION
	//Translate and rote links, need two matrices to save temp result in
	ob::ProjectionMatrix::Matrix tempResult1(4,4);
	ob::ProjectionMatrix::Matrix tempResult2(4,4);
	
	//Need to know how many triangles we have made to get correct triangle number we with PQP
	int numberOfTriangles = environmentTriangles;
	
	//T0-1
	PQP_Model m1;
	numberOfTriangles = createPQPModel(m1, A0, numberOfTriangles, 1);
	
	//T0-2 get saved in tempResult1
	PQP_Model m2;
	numberOfTriangles = createPQPModel(m2, A1, numberOfTriangles, 2);
	
	//T0-3
	DH::mul(tempResult2, A1, A2);
	PQP_Model m3;
	numberOfTriangles = createPQPModel(m3, tempResult2, numberOfTriangles, 3);
	
	//T0-4
	DH::mul(tempResult1, tempResult2, A3);
	PQP_Model m4;
	numberOfTriangles = createPQPModel(m4, tempResult1, numberOfTriangles, 4);
	
	//T0-5
	DH::mul(tempResult2, tempResult1, A4);
	PQP_Model m5;
	numberOfTriangles = createPQPModel(m5, tempResult2, numberOfTriangles, 5);
	
	//T0-6
	DH::mul(tempResult1, tempResult2, A5);
	PQP_Model m6;
	numberOfTriangles = createPQPModel(m6, tempResult1, numberOfTriangles, 6);
  	
  	int selfCollision = checkForSelfCollision(m1, m2, m3, m4, m5, m6);

  	int envCollision = checkForCollisionWithEnvironment((PQP_Model&)environment, m1, m2, m3, m4, m5, m6);

	if(selfCollision == 1){
		//std::cout << "Self collision detected" << std::endl;
		return false;
	}
	else if(envCollision == 1){
		//std::cout << "Collision with environment detected" << std::endl;
		return false;
		
	}
	else{
	//std::cout << "No collision detected" << std::endl;
		return true;
	}
}

int myStateValidityCheckerClass::createPQPModel(PQP_Model &model, ob::ProjectionMatrix::Matrix orientation, int numberOfTriangles, int linkNumber) const{
	ob::ProjectionMatrix::Matrix tempResult(4,4);
	PQP_REAL cornerPoints[36][3];

	for(int i=0; i<36; i++){
		switch(linkNumber){
			case 1:
				DH::mul(tempResult, orientation, linkOne[i]);
				break;
			case 2:
				DH::mul(tempResult, orientation, linkTwo[i]);
				break;
			case 3:
				DH::mul(tempResult, orientation, linkThree[i]);
				break;		
			case 4:
				DH::mul(tempResult, orientation, linkFour[i]);
				break;
			case 5:
				DH::mul(tempResult, orientation, linkFive[i]);
				break;
			case 6:
				DH::mul(tempResult, orientation, linkSix[i]);
				break;
		}
		cornerPoints[i][0] = tempResult(0,3);
		cornerPoints[i][1] = tempResult(1,3);
		cornerPoints[i][2] = tempResult(2,3);
	}
	model.BeginModel();
	//Three points per triangle
	for(int j=0; j<36; j=j+3){
		numberOfTriangles = numberOfTriangles +1;
		model.AddTri(cornerPoints[j], cornerPoints[j+1], cornerPoints[j+2], numberOfTriangles);
	}
	model.EndModel();
	return numberOfTriangles;
}

int myStateValidityCheckerClass::checkForSelfCollision(PQP_Model &m1, PQP_Model &m2, PQP_Model &m3, PQP_Model &m4, PQP_Model &m5, PQP_Model &m6) const{
	//Position and rotation is already calculated into the PQP model so we can use an identity matrix for rotation
	PQP_REAL R1[3][3];
	
	for(int i=0; i<3; i++){
		for(int j=0; j<3; j++){
			if(i==j){
				R1[i][j] = 1;
			}
			else{
				R1[i][j] = 0;
			}
		}	
	}
	
	PQP_REAL T1[3]; 
	
	T1[0] = 0;
	T1[1] = 0;
	T1[2] = 0;
	
	//Check for collisions
	PQP_CollideResult cres12;
	PQP_Collide(&cres12, R1, T1, &m1, R1, T1, &m2);
	PQP_CollideResult cres13;
	PQP_Collide(&cres13, R1, T1, &m1, R1, T1, &m3);
	PQP_CollideResult cres14;
	PQP_Collide(&cres14, R1, T1, &m1, R1, T1, &m4);
	PQP_CollideResult cres15;
	PQP_Collide(&cres15, R1, T1, &m1, R1, T1, &m5);
	PQP_CollideResult cres16;
	PQP_Collide(&cres16, R1, T1, &m1, R1, T1, &m6);
	
	PQP_CollideResult cres23;
	PQP_Collide(&cres23, R1, T1, &m2, R1, T1, &m3);
	PQP_CollideResult cres24;
	PQP_Collide(&cres24, R1, T1, &m2, R1, T1, &m4);
	PQP_CollideResult cres25;
	PQP_Collide(&cres25, R1, T1, &m2, R1, T1, &m5);
	PQP_CollideResult cres26;
	PQP_Collide(&cres26, R1, T1, &m2, R1, T1, &m6);
	
	PQP_CollideResult cres34;
	PQP_Collide(&cres34, R1, T1, &m3, R1, T1, &m4);
	PQP_CollideResult cres35;
	PQP_Collide(&cres35, R1, T1, &m3, R1, T1, &m5);
	PQP_CollideResult cres36;
	PQP_Collide(&cres36, R1, T1, &m3, R1, T1, &m6);
	
	PQP_CollideResult cres45;
	PQP_Collide(&cres45, R1, T1, &m4, R1, T1, &m5);
	PQP_CollideResult cres46;
	PQP_Collide(&cres46, R1, T1, &m4, R1, T1, &m6);
	
	PQP_CollideResult cres56;
	PQP_Collide(&cres56, R1, T1, &m5, R1, T1, &m6);
	
	int colliding = 0;
	int colliding12 = cres12.Colliding();
	int colliding13 = cres13.Colliding();
	int colliding14 = cres14.Colliding();
	int colliding15 = cres15.Colliding();
	int colliding16 = cres16.Colliding();
	
	int colliding23 = cres23.Colliding();
	int colliding24 = cres24.Colliding();
	int colliding25 = cres25.Colliding();
	int colliding26 = cres26.Colliding();
	
	int colliding34 = cres34.Colliding();
	int colliding35 = cres35.Colliding();
	int colliding36 = cres36.Colliding();
	
	int colliding45 = cres45.Colliding();
	int colliding46 = cres46.Colliding();
	
	int colliding56 = cres56.Colliding();
	
	if(colliding12 == 1 || colliding13 == 1 || colliding14 ==1 || colliding15 == 1 || colliding16 ==1 || colliding23 == 1 || colliding24 == 1 || colliding25 == 1 || colliding26 ==1 || colliding34 ==1 || colliding35 ==1 || colliding36 ==1 || colliding45 ==1 || colliding46 ==1 || colliding56 ==1){ 
		colliding = 1;
	}
	else{
		colliding = 0;
	}
	
	return colliding;
}

int myStateValidityCheckerClass::checkForCollisionWithEnvironment(PQP_Model &envCheck, PQP_Model &m1, PQP_Model &m2, PQP_Model &m3, PQP_Model &m4, PQP_Model &m5, PQP_Model &m6) const{

	PQP_REAL R1[3][3];
	
	for(int i=0; i<3; i++){
		for(int j=0; j<3; j++){
			if(i==j){
				R1[i][j] = 1;
			}
			else{
				R1[i][j] = 0;
			}
		}	
	}
	
	PQP_REAL T1[3];
	
	T1[0] = 0;
	T1[1] = 0;
	T1[2] = 0;
	
	//Check for collisions
	int colliding = 0;
	
	PQP_CollideResult envCres1;
	PQP_Collide(&envCres1, R1, T1, &envCheck, R1, T1, &m1);	
	int colliding1 = envCres1.Colliding();

	PQP_CollideResult envCres2;
	PQP_Collide(&envCres2, R1, T1, &envCheck, R1, T1, &m2);	
	int colliding2 = envCres2.Colliding();
	
	PQP_CollideResult envCres3;
	PQP_Collide(&envCres3, R1, T1, &envCheck, R1, T1, &m3);	
	int colliding3 = envCres3.Colliding();
	
	PQP_CollideResult envCres4;
	PQP_Collide(&envCres4, R1, T1, &envCheck, R1, T1, &m4);	
	int colliding4 = envCres4.Colliding();
	
	PQP_CollideResult envCres5;
	PQP_Collide(&envCres5, R1, T1, &envCheck, R1, T1, &m5);	
	int colliding5 = envCres5.Colliding();
	
	PQP_CollideResult envCres6;
	PQP_Collide(&envCres6, R1, T1, &envCheck, R1, T1, &m6);	
	int colliding6 = envCres6.Colliding();
	
	if(colliding1 == 1 || colliding2 == 1 || colliding3 ==1 || colliding4 ==1 || colliding5 ==1 || colliding6 ==1){
		colliding = 1;
	}
	//if no collision is found colliding == 0
	return colliding;
}

void myStateValidityCheckerClass::loadCornerPoints(){
	importLink("irb/Robot0", link1, 1);
	importLink("irb/Robot1", link2, 2);
	importLink("irb/Robot2", link3, 3);
	importLink("irb/Robot3", link4, 4);
	importLink("irb/Robot4", link5, 5);
	importLink("irb/Robot5", link6, 6);
}
	

void myStateValidityCheckerClass::importLink(const char *linkFil, std::vector<float > &linkVector, int linkNumber){
	std::cout << "Importing link "<< linkNumber << std::endl;
	std::fstream file;
	file.open(linkFil);
	char line[50];
	
	while(file.getline(line, 50)){

		float pts[9];
	  	sscanf(line, "(%f,%f,%f) (%f,%f,%f) (%f,%f,%f)", 
		&pts[0], &pts[1], &pts[2], 
		&pts[3], &pts[4], &pts[5],
		&pts[6], &pts[7], &pts[8]);		
		
		for(int i=0; i<9;i++){
			linkVector.push_back(pts[i]);
		}
	}
	file.close();

	ob::ProjectionMatrix::Matrix ident(4,4);
	DH::identity(ident);
	
	int k=-1;
	for(int i=0; i<108;i++){
		switch(linkNumber){
			case 1:
			if(i%3 == 0){
				k++;
				linkOne[k] = ident;
			}
			linkOne[k](i%3,3) = linkVector[i];
			break;
			
			case 2:
			if(i%3 == 0){
				k++;
				linkTwo[k] = ident;
			}
			linkTwo[k](i%3,3) = linkVector[i];	
			break;
			
			case 3:
			if(i%3 == 0){
				k++;
				linkThree[k] = ident;
			}
			linkThree[k](i%3,3) = linkVector[i];	
			break;
			
			case 4:
			if(i%3 == 0){
				k++;
				linkFour[k] = ident;
			}
			linkFour[k](i%3,3) = linkVector[i];	
			break;
			
			case 5:
			if(i%3 == 0){
				k++;
				linkFive[k] = ident;
			}
			linkFive[k](i%3,3) = linkVector[i];
			break;
			
			case 6:
			if(i%3 == 0){
				k++;
				linkSix[k] = ident;
			}
			linkSix[k](i%3,3) = linkVector[i];				
		}
	}
	
	std::cout << "Link " << linkNumber << " imported" << std::endl;
}

int myStateValidityCheckerClass::importEnvironment(const char *linkFil, PQP_Model &env, int numberOfTriangles)
{
	
	std::fstream fileEnv;
	fileEnv.open(linkFil);
	char lineEnv[500];
	char pts0[50];
	
	while(fileEnv.getline(lineEnv, 500)){
	  	sscanf(lineEnv, "%s", 
		pts0);
	}
	fileEnv.close();
	
	std::cout << "Importing Environment " << pts0 << std::endl;

	std::fstream file;
	file.open(pts0);
	char line[500];
	
	std::vector<float> envVector;
	
	while(file.getline(line, 500)){
		float pts[9];
	  	sscanf(line, "%f %f %f %f %f %f %f %f %f", 
		&pts[0], &pts[1], &pts[2], 
		&pts[3], &pts[4], &pts[5],
		&pts[6], &pts[7], &pts[8]);		
		for(int i=0; i<9;i++){
			envVector.push_back(pts[i]);
		}
	}
	file.close();
	
	PQP_REAL cornerPoints[envVector.size()/3][3];
	int h= 0;
	//Environment made in Blender needs to be multiplied by some factor in order to be correct
	int xMul = 140;
	int yMul = 115;
	int zMul = 200;
	
	for(int k=0; k<(int)envVector.size(); k=k+3){
		//The robotmodel in Blender is shifted 100 units from the model used in all directions.
		//The environment have to be shifted the same amount of units in order to be placed correctly
		cornerPoints[h][0] = envVector[k]*xMul -100;
		cornerPoints[h][1] = envVector[k+1]*yMul -100;
		cornerPoints[h][2] = envVector[k+2]*zMul - 100;
		h++;
	}
	
	env.BeginModel();
	
	//Three points per triangle
	for(int j=0; j<h; j=j+3){
		env.AddTri(cornerPoints[j], cornerPoints[j+1], cornerPoints[j+2], numberOfTriangles);
		numberOfTriangles = numberOfTriangles +1;
	}
	env.EndModel();
	std::cout << "Environment Imported with " << numberOfTriangles << " triangles" << std::endl;
	return numberOfTriangles;
	
}