 #include "planning.h"
#include <string.h>

void planning::planWithSS()
{
	std::cout << "***************** PLANNING WITH SIMPLE SETUP STARTED **********************" << std::endl;
	//Create statespace
	ob::StateSpacePtr R6(new ob::RealVectorStateSpace(6));
	R6->setName("State Space, R6");
	 
	//set bounds for each link in R6
	ob::RealVectorBounds bounds(6);
	bounds.setLow(0, -PI); 	//360 degrees
	bounds.setHigh(0, +PI); 	//360 degrees
	
	bounds.setLow(1, -(10*PI)/18);	//200degrees
	bounds.setHigh(1, (10*PI)/18);	//200degrees
	
	bounds.setLow(2, -(14*PI)/18);	//280 degrees
	bounds.setHigh(2, (14*PI)/18);	//280 degrees
	
	bounds.setLow(3, -(20*PI)/18);	//Unlimited(400 default)
	bounds.setHigh(3, (20*PI)/18);	//Unlimited
	
	bounds.setLow(4, -(4*PI)/6);		//240 degrees
	bounds.setHigh(4, (4*PI)/6);	//240 degrees
	
	bounds.setLow(5, -(40*PI)/18);	//Unlimited(800 default)
	bounds.setHigh(5, (40*PI)/18);	//Unlimited
		
	R6->as<ob::RealVectorStateSpace>()->setBounds(bounds);
	
	//Construct space information from R6
	og::SimpleSetup ss(R6);
	
	//Set state validity checking
	ss.setStateValidityChecker(ob::StateValidityCheckerPtr(new myStateValidityCheckerClass(ss.getSpaceInformation())));
	
	//create randoms start state
	ob::ScopedState<> start(R6);
	setPosition(start, "irb/start");
	//start.random();
	
	//set goal state
	ob::ScopedState<> goal(R6);
	setPosition(goal, "irb/goal");
	//goal.random();
	
	//Set start and goal
	ss.setStartAndGoalStates(start, goal);

	//Create a planner
	ob::PlannerPtr planner(new og::RRT(ss.getSpaceInformation()));
	ss.setPlanner(planner);
	
	//Attempt to solve problem. Time in solved is time allowed to be used for finding path
	 bool solved = ss.solve(7);
	
	//Check if the solution is approximate
	bool approx = ss.getGoal()->isApproximate();
	
	if (solved && !approx){
		std::cout << "******* VALID PATH FOUND *******" << std::endl;
		//get path and goal representation
		og::PathGeometric pathbeforess = ss.getSolutionPath();
		std::cout << "******** Before simplification ********" << std::endl;
		pathbeforess.print(std::cout);

		ss.simplifySolution();

		og::PathGeometric path = ss.getSolutionPath();
				
		path.interpolate(20);
		std::cout << "******** After Simplification ********" << std::endl;
		path.print(std::cout);
		
		//Calculate DH transformations for current state
		ob::ProjectionMatrix::Matrix A1 = DH::makeAmatrix(352, goal[0] + 0, 70, -PI/2);
		ob::ProjectionMatrix::Matrix A2 = DH::makeAmatrix(0, goal[1] + (-PI/2), 360, 0);
		ob::ProjectionMatrix::Matrix A3 = DH::makeAmatrix(0, goal[2] + 0, 0, -PI/2);
		ob::ProjectionMatrix::Matrix A4 = DH::makeAmatrix(445, goal[3] + 0, 0, PI/2);
		ob::ProjectionMatrix::Matrix A5 = DH::makeAmatrix(0, goal[4] + 0, 0, -PI/2);
		ob::ProjectionMatrix::Matrix A6 = DH::makeAmatrix(0, goal[5] + PI, 0, 0);
	
		//Use forward kinematics to find final position and rotation
		ob::ProjectionMatrix::Matrix H(4,4);
		DH::fullDHIRB140(H, A1, A2, A3, A4, A5, A6); 
	
		planning::writePathToFileWithSS(path);
	
		std::cout << "*********** GOAL DH ***********" << std::endl;
		for(int a=0; a<4;a++)
		{
			for(int b=0; b<4; b++)
			{
				std::cout << H(a,b) << "\t";
			}
			std::cout << std::endl;
		}	
		std::cout << "******************************" << std::endl;

		
	}
	else{
		std::cout << "No solutiuon found" << std::endl;
		if(approx){
			std::cout << "An approximate solution was found, but is not wanted" << std::endl;
		}
	}
}

void planning::writePathToFileWithSS(og::PathGeometric solutionPath)
{
	std::cout << "WRITING PATH TO FILE" << std::endl;
	//print path to file
	std::ofstream pathFileWrite;
	pathFileWrite.open("dummy.txt");
	solutionPath.print(pathFileWrite<<std::cout);
	pathFileWrite.close();
	
	//remove unwanted information so only joint angles are left
	//The dummy file is only a step in getting the full path.
	std::ifstream pathFileRead;
	pathFileRead.open("dummy.txt");
	
	std::ofstream purePath;
	purePath.open("validPath.txt");
	
	std::ofstream pureElbowPath;
	pureElbowPath.open("validPathElbow.txt");
	
	if(!pathFileRead.is_open() || !purePath.is_open() || !pureElbowPath.is_open()){
		std::cout << "ERROR: One of the files needed was not opened correctly" << std::endl;
	}
	char line[100];
	
	//First line is not interesting as it only says how many states are in the path so we skip this line
	char dummyLine[100];
	pathFileRead.getline(dummyLine, 100);
	
	while(pathFileRead.getline(line, 100)){
		float pts[6];
	  	sscanf(line, "RealVectorState [%f %f %f %f %f %f]", 
		&pts[0], &pts[1], &pts[2], 
		&pts[3], &pts[4], &pts[5]
		);
		if(strlen(line) == 0){
			break;
		}
		purePath << pts[0] << " " << pts[1] << " " << pts[2] << " " << pts[3] << " " << pts[4] << " " << pts[5] << "\n";
		pureElbowPath << pts[0] << " " << pts[1] << " " << pts[2] << "\n";
	}
	pathFileRead.close();
	purePath.close();
	pureElbowPath.close();
	
	std::cout << "PATH WRITTEN TO FILE" << std::endl;
}

void planning::setPosition(ob::ScopedState<> &position, const char *linkFil){
	std::cout << "Setting " << linkFil << " position " << std::endl;
	std::fstream file;
	file.open(linkFil);
	char line[50];
	float pts[9];
	while(file.getline(line, 50)){
	  	sscanf(line, "%f %f %f %f %f %f", 
		&pts[0], &pts[1], &pts[2], 
		&pts[3], &pts[4], &pts[5]);
	}
	file.close();
	
	for(int i=0; i<6; i++){
		position[i]  = pts[i];
	}
}

void planning::planWithBenchmark()
{
	std::cout << "***************** PLANNING WITH SIMPLE SETUP STARTED **********************" << std::endl;
	//Create statespace
	ob::StateSpacePtr R6(new ob::RealVectorStateSpace(6));
	R6->setName("State Space, R6");
	 
	//set bounds for each link in R6
	ob::RealVectorBounds bounds(6);
	bounds.setLow(0, -PI); 	//360 degrees
	bounds.setHigh(0, PI); 	//360 degrees
	
	bounds.setLow(1, -(10*PI)/18);	//200degrees
	bounds.setHigh(1, (10*PI)/18);	//200degrees
	
	bounds.setLow(2, -(14*PI)/18);	//280 degrees
	bounds.setHigh(2, (14*PI)/18);	//280 degrees
	
	bounds.setLow(3, -(20*PI)/18);	//Unlimited(400 default)
	bounds.setHigh(3, (20*PI)/18);	//Unlimited
	
	bounds.setLow(4, -(4*PI)/6);		//240 degrees
	bounds.setHigh(4, (4*PI)/6);	//240 degrees
	
	bounds.setLow(5, -(40*PI)/18);	//Unlimited(800 default)
	bounds.setHigh(5, (40*PI)/18);	//Unlimited
		
	R6->as<ob::RealVectorStateSpace>()->setBounds(bounds);
	
	//Construct space information from R6
	og::SimpleSetup ss(R6);
	
	//Set state validity checking
	ss.setStateValidityChecker(ob::StateValidityCheckerPtr(new myStateValidityCheckerClass(ss.getSpaceInformation())));
	
	//create randoms start state
	ob::ScopedState<> start(R6);
	setPosition(start, "irb/start");
	
	//set goal state
	ob::ScopedState<> goal(R6);
	setPosition(goal, "irb/goal");

	//Set start and goal
	ss.setStartAndGoalStates(start, goal);

	//Create a planner
	ob::PlannerPtr planner(new og::RRT(ss.getSpaceInformation()));
	ss.setPlanner(planner);
	
	ompl::tools::Benchmark b(ss, "benchmarking");
	b.addPlanner(ob::PlannerPtr(new og::KPIECE1(ss.getSpaceInformation())));
	b.addPlanner(ob::PlannerPtr(new og::RRT(ss.getSpaceInformation())));
	b.addPlanner(ob::PlannerPtr(new og::RRTConnect(ss.getSpaceInformation())));
	b.addPlanner(ob::PlannerPtr(new og::LazyRRT(ss.getSpaceInformation())));
	b.addPlanner(ob::PlannerPtr(new og::LBKPIECE1(ss.getSpaceInformation())));
	
	ompl::tools::Benchmark::Request req;
	req.maxTime = 5.0;
	req.maxMem = 100.0;
	req.runCount = 50;
	req.displayProgress = true;
	b.benchmark(req);
	
	b.saveResultsToFile("benchResult");
}